/*
 *  Copyright (C) 2013, Northwestern University.
 *  See COPYRIGHT notice COPYING in top-level directory.
 *
 *  $Id: gcrm_param_init.c 1858 2013-06-12 04:49:45Z wkliao $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <mpi.h>

#include "gcrm.h"
#include "gio.h"
#include "ZGrd_clusdet.h"
#include "ZGrd_output.h"
#include "ZGrd_params.h"
#include "ZGrd_params_tracer.h"
#include "ZGrd_params_vertical.h"
#include "ZGrd_params_time.h"
#include "ZGrd_vars_diagnostic.h"
#include "ZGrd_vars_prognostic.h"
#include "util.h"

/*----< parallel_load_inputs() >----------------------------------------------*/
/* from parallel/parallel_params.F90 */
static
void parallel_load_inputs(char            *fname,
                          gcrm_parameters *gcrm_param)
{
    if (gcrm_param->rnk_wrld == 0) { /* only root reads from the file */
        char line[256], *value, *key;
        FILE *fp;

        OPEN_FILE(fname)

        while (fgets(line, 256, fp) != NULL) {
            GET_LINE_TOKEN_PAIR

            /* from parallel_load_inputs() called in ZGrd_main */
            if (!strcasecmp(key, "io_procs")) {
                gcrm_param->npe_io = atoi(value);
                break;
            }
        }
        fclose(fp);
    }
    BCAST_INT(gcrm_param->npe_io)
}

/*----< initialize_ZGrd() >---------------------------------------------------*/
/* initialize_ZGrd() is defined in ZGrd_initialize.F90 */
static
void initialize_ZGrd(char                *in_fname,
                     gcrm_parameters     *gcrm_param,
                     gio_parameters      *gio_param)
{
    MODULE_grid_params *grid_params = &gcrm_param->grid_params;

    gcrm_param->enable_physics=1;

    if (!strncasecmp(gcrm_param->zgrd_params.physics_mode, "disabled", 7))
        gcrm_param->enable_physics = 0;  /* not disabled */

    /* set subdomains. partition global grid and assign pieces to processes
       in grid/grid_subdomain.F90 */
    initialize_subdomain(grid_params, &gcrm_param->grid_subdomain);

    /* set tree data structure. each process builds its portion of the global
       grid.  grid/grid_connectivity.F90 */
    initialize_grid_connectivity(grid_params, "world");

    /* set parallel communication for ghost cell updates of 2D-array data
       structure */
    initialize_wrap(&gcrm_param->wrap_data,
                    grid_params,
                    "world","ZGrd",
                    grid_params->level_max,
                    grid_params->sbdmn[grid_params->level_max].sbdmn_iota,
                    grid_params->sbdmn[grid_params->level_max].lst,
                    0);

    /* set grid metrics */
    initialize_grid_metrics(grid_params,
                            &gcrm_param->grid_subdomain,
                            &gcrm_param->grid_metrics,
                            &gcrm_param->wrap_data,
                            "ZGrd",
                            gcrm_param->zgrd_params.grid_point_select,
                            NULL);

    /* call initialize_params_vertical() in ZGrd_params_vertical.F90 */
    initialize_params_vertical(&gcrm_param->vertical);

    /* call initialize_params_time() in ZGrd_params_time.F90 */
    initialize_params_time(in_fname, &gcrm_param->time,
                           gcrm_param->grid_params.level_max);

    init_MODULE_ZGrd_params_tracer(&gcrm_param->tracer,
                                   gcrm_param->enable_physics);

    /* this must be called after initialize_params_time(), as time.nprog
     * and time.ntend are defined there */
    init_MODULE_ZGrd_vars_prognostic(&gcrm_param->vars_prog,
                                     gcrm_param->grid_params.im,
                                     gcrm_param->grid_params.jm,
                                     gcrm_param->vertical.km,
                                     gcrm_param->time.nprog,
                                     gcrm_param->grid_params.nsdm,
                                     gcrm_param->time.ntend,
                                     gcrm_param->tracer.wtrm,
                                     gcrm_param->tracer.trcm);

    initialize_vars_prognostic(&gcrm_param->vars_prog,
                                     gcrm_param->grid_params.im,
                                     gcrm_param->grid_params.jm,
                                     gcrm_param->vertical.km,
                                     gcrm_param->time.nprog,
                                     gcrm_param->grid_params.nsdm,
                                     gcrm_param->time.ntend,
                                     gcrm_param->tracer.wtrm,
                                     gcrm_param->tracer.trcm);

    init_MODULE_ZGrd_vars_diagnostic(&gcrm_param->vars_diag,
                                     gcrm_param->grid_params.im,
                                     gcrm_param->grid_params.jm,
                                     gcrm_param->vertical.km,
                                     gcrm_param->grid_params.nsdm,
                                     gcrm_param->tracer.wtrm,
                                     gcrm_param->tracer.trcm,
                                     gcrm_param->grid_params.edgm,
                                     gcrm_param->grid_params.crnm);

    initialize_vars_diagnostic(&gcrm_param->vars_diag,
                                     gcrm_param->grid_params.im,
                                     gcrm_param->grid_params.jm,
                                     gcrm_param->vertical.km,
                                     gcrm_param->grid_params.nsdm,
                                     gcrm_param->tracer.wtrm,
                                     gcrm_param->tracer.trcm,
                                     gcrm_param->grid_params.edgm,
                                     gcrm_param->grid_params.crnm);

    if (gcrm_param->enable_physics)
        init_MODULE_phys_vars_diag_out(&gcrm_param->phys_vars,
                                     gcrm_param->grid_params.nsdm,
                                     gcrm_param->vertical.km,
                                     gcrm_param->grid_params.jm,
                                     gcrm_param->grid_params.im);

    /* pass GCRM parameters to GIO */
    gio_init(gcrm_param, gio_param);

    /* initialize_gio_api() is called in ZGrd_register.F90
     * It sets contents of all grid variables and register all grid and
     * non-grid variables to the gio library */
    initialize_gio_api(gio_param,
                       &gcrm_param->tracer,
                       &gcrm_param->grid_params,
                       &gcrm_param->grid_metrics,
                       &gcrm_param->vertical,
                       &gcrm_param->vars_diag,
                       &gcrm_param->vars_prog,
                       &gcrm_param->clusdet,
                       &gcrm_param->time,
                       &gcrm_param->phys_vars,
                       gcrm_param->enable_physics);
}

/*----< pseudo_computation() >------------------------------------------------*/
void pseudo_computation(gcrm_parameters *gcrm_param)
{
    /* just some dummy computation of random number generation */
    randomize_MODULE_ZGrd_clusdet(&gcrm_param->clusdet,
                                  gcrm_param->grid_params.im,
                                  gcrm_param->grid_params.jm,
                                  gcrm_param->vertical.km,
                                  gcrm_param->grid_params.nsdm,
                                  gcrm_param->grid_params.edgm,
                                  gcrm_param->grid_params.crnm,
                                  gcrm_param->npe_wrld);

    randomize_ZGrd_vars_prognostic(&gcrm_param->vars_prog,
                                   gcrm_param->grid_params.im,
                                   gcrm_param->grid_params.jm,
                                   gcrm_param->vertical.km,
                                   gcrm_param->time.nprog,
                                   gcrm_param->grid_params.nsdm,
                                   gcrm_param->time.ntend,
                                   gcrm_param->tracer.wtrm,
                                   gcrm_param->tracer.trcm);

    randomize_ZGrd_vars_diagnostic(&gcrm_param->vars_diag,
                                   gcrm_param->grid_params.im,
                                   gcrm_param->grid_params.jm,
                                   gcrm_param->vertical.km,
                                   gcrm_param->grid_params.nsdm,
                                   gcrm_param->tracer.wtrm,
                                   gcrm_param->tracer.trcm,
                                   gcrm_param->grid_params.edgm,
                                   gcrm_param->grid_params.crnm);

    if (gcrm_param->enable_physics)
        randomize_phys_vars_diag_out(&gcrm_param->phys_vars,
                                     gcrm_param->grid_params.nsdm,
                                     gcrm_param->vertical.km,
                                     gcrm_param->grid_params.jm,
                                     gcrm_param->grid_params.im);
}

/*----< gcrm_param_init() >---------------------------------------------------*/
/* this subroutine is based on ZGrd_main.F90  */
int gcrm_param_init(char                *in_fname,
                    gcrm_parameters     *gcrm_param,
                    gio_parameters      *gio_param)
{
    /* zero out the entire contents of gcrm_param object */
    memset(gcrm_param, 0, sizeof(gcrm_param));

    MPI_Comm_rank(MPI_COMM_WORLD, &gcrm_param->rnk_wrld);
    MPI_Comm_size(MPI_COMM_WORLD, &gcrm_param->npe_wrld);

    /* set random number seed */
    srandom(gcrm_param->rnk_wrld);

    /* All functions called below with prefix name init_MODULE_ correspond to
     * the F90 modules that define array variables with static sizes and
     * dimensions.
     * In this C version, definition for all static array variables have been
     * changed to dynamic memory allocation, and initialized to zeros.
     * Their malloc must be done before their initializations.
     */

    /* File: grid/grid_params.F90, Fortran MODULE grid_params */
    init_MODULE_grid_params(&gcrm_param->grid_params, in_fname);

    /* File: grid/grid_subdomain.F90 Fortran MODULE grid_subdomain */
    init_MODULE_grid_subdomain(&gcrm_param->grid_subdomain,
                               gcrm_param->grid_params.nsdm);

    /* File: grid/grid_metrics.F90 Fortran MODULE grid_metrics */
    init_MODULE_grid_metrics(&gcrm_param->grid_metrics,
                             gcrm_param->grid_params.im,
                             gcrm_param->grid_params.jm,
                             gcrm_param->grid_params.nsdm,
                             gcrm_param->grid_params.edgm,
                             gcrm_param->grid_params.crnm);

    /* File: ZGrd/ZGrd_output.F90, Fortran MODULE ZGrd_output */
    init_MODULE_ZGrd_output(&gcrm_param->zgrd_output);

    /* array memory allocation in ZGrd_clusdet.F90 */
    init_MODULE_ZGrd_clusdet(&gcrm_param->clusdet,
                             gcrm_param->grid_params.im,
                             gcrm_param->grid_params.jm,
                             gcrm_param->vertical.km,
                             gcrm_param->grid_params.nsdm,
                             gcrm_param->grid_params.edgm,
                             gcrm_param->grid_params.crnm,
                             gcrm_param->npe_wrld);

    init_MODULE_wrap_data(&gcrm_param->wrap_data);

    /* Now below are the functions called in main() in ZGrd_main.F90 */
    //     call gin_load_inputs() in utilities/ginput.F90

    /* parallel_load_inputs() is in parallel/parallel_params.F90
     * It only reads io_procs from the input parameter file */
    parallel_load_inputs(in_fname, gcrm_param);

    /* io_load_inputs() is defined in ZGrd_output.F90
     * It reads input parameter file (e.g. zgrd.in) */
    io_load_inputs(in_fname, &gcrm_param->zgrd_output);

    /* ZGrd_load_inputs() is defined in ZGrd_params.F90 */
    ZGrd_load_inputs(in_fname, &gcrm_param->zgrd_params);

    /* clusdet_initialize() is defined in ZGrd_clusdet.F90
     * It only reads clusdet_enable from input parameter file */
    clusdet_initialize(in_fname, &gcrm_param->clusdet);

    /* initialize_ZGrd() is defined in ZGrd_initialize.F90 */
    initialize_ZGrd(in_fname, gcrm_param, gio_param);

    return 1;
}

/*----< gcrm_param_finalize() >-----------------------------------------------*/
int gcrm_param_finalize(gcrm_parameters *gcrm_param,
                        gio_parameters  *gio_param)
{
    if (gcrm_param->enable_physics)
        finalize_MODULE_phys_vars_diag_out(&gcrm_param->phys_vars);

    finalize_MODULE_ZGrd_vars_diagnostic(&gcrm_param->vars_diag);
    finalize_MODULE_ZGrd_vars_prognostic(&gcrm_param->vars_prog);
    finalize_MODULE_wrap_data(&gcrm_param->wrap_data);
    finalize_MODULE_ZGrd_clusdet(&gcrm_param->clusdet);
    finalize_MODULE_grid_metrics(&gcrm_param->grid_metrics);
    finalize_MODULE_grid_subdomain(&gcrm_param->grid_params,
                                   &gcrm_param->grid_subdomain);
    finalize_MODULE_grid_params(&gcrm_param->grid_params);

    return 1;
}
